# -*- coding: utf-8 -*-
# Author: Subhasis Ray <ray dot subhasis at gmail dot com>
# Created: 2021-03-16 12:20 PM

import sys
import logging
import yaml
from PyQt5 import (
    QtWidgets as qw,
)

from argos_track.track import ArgosTracker


app = qw.QApplication(sys.argv)
win = ArgosTracker()
win.setMinimumSize(800, 600)
win.setWindowTitle('Argos - track animals in video')
win.showMaximized()
app.aboutToQuit.connect(win.cleanup)
win.show()
sys.exit(app.exec_())


# -*- coding: utf-8 -*-
# Author: Subhasis Ray <ray dot subhasis at gmail dot com>
# Created: 2020-06-01 11:50 PM

"""
===========================================
SORT algorithm for multiple object tracking
===========================================

Implemented with OpenCV's builtin Kalman filter.

This algorithm was originally developed and published by Alex Bewly [2]_.

Original implementation of SORT_ and DeepSORT_

References
----------
.. [2] Bewley, A., Ge, Z., Ott, L., Ramos, F. & Upcroft, B. Simple
   Online and Realtime Tracking. 2016 IEEE International Conference on
   Image Processing (ICIP) 34643468 (2016)
   doi:10.1109/ICIP.2016.7533003.

.. _SORT: https://github.com/abewley/sort

.. _DeepSORT: https://github.com/nwojke/deep_sort.

"""
import logging
import numpy as np
import cv2
from PyQt5 import (
    QtCore as qc)

import argos.constants
from argos import utility as au
from argos.utility import match_bboxes

settings = au.init()


class KalmanTracker(object):
    """This class tries to improve performance over SORT or DeepSORT by using
    opencv's builtin Kalman Filter. OpenCV being written in C/C++ it outperforms
    the Python code in DeepSORT or filterpy (used in SORT).

    In my test, the predict step in OpenCV takes

    2.78 µs ± 14.7 ns per loop (mean ± std. dev. of 7 runs, 100000 loops each)

    compared to DeepSORT taking

    45.7 µs ± 1.24 µs per loop (mean ± std. dev. of 7 runs, 10000 loops each)
    """
    NDIM = 4
    DT = 1.0

    def __init__(self, bbox, track_id, min_hits=3, max_age=10, deepsort=False):
        """bbox is in xywh format and converted to xyrh format"""
        super(KalmanTracker, self).__init__()
        self.tid = track_id
        self.hits = 1
        self.min_hits = min_hits
        self.features = []
        self.time_since_update = 0
        self.n_init = min_hits
        self.max_age = max_age
        self._std_weight_pos = 1.0 / 20
        self._std_weight_vel = 1.0 / 160
        # flag to switch between fixed covariances like SORT vs
        # measurement-based covariance like DeepSORT
        self.cov_deepsort = deepsort
        self.filter = cv2.KalmanFilter(dynamParams=2 * self.NDIM,
                                       measureParams=self.NDIM, type=cv2.CV_64F)
        # Borrowing ideas from SORT/DeepSORT
        # Measurement marix H
        self.filter.measurementMatrix = np.eye(self.NDIM, 2 * self.NDIM)

        # This is state transition matrix F
        self.filter.transitionMatrix = np.eye(2 * self.NDIM, 2 * self.NDIM)
        for ii in range(self.NDIM):
            self.filter.transitionMatrix[ii, ii + self.NDIM] = self.DT
        # NOTE state covariance matrix (P) is initialized as a function of
        # measured height in DeepSORT, but constant in SORT.
        if self.cov_deepsort:
            error_cov = [2 * self._std_weight_pos * bbox[3],
                         2 * self._std_weight_pos * bbox[3],
                         1e-2,
                         2 * self._std_weight_pos * bbox[3],
                         10 * self._std_weight_vel * bbox[3],
                         10 * self._std_weight_vel * bbox[3],
                         1e-5,
                         10 * self._std_weight_vel * bbox[3]]
            self.filter.errorCovPost = np.diag(np.square(error_cov))
        else:
            self.filter.errorCovPost = np.eye(2 * self.NDIM, dtype=float) * 10.0
            self.filter.errorCovPost[self.NDIM:,
            self.NDIM:] *= 1000.0  # High uncertainty for velocity at first

        # NOTE process noise covariance matrix (Q) [here motion covariance] is
        # computed as a function of mean height in DeepSORT, but constant
        # in SORT

        if self.cov_deepsort:
            proc_cov = [self._std_weight_pos * bbox[3],
                        self._std_weight_pos * bbox[3],
                        1e-2,
                        self._std_weight_pos * bbox[3],
                        self._std_weight_vel * bbox[3],
                        self._std_weight_vel * bbox[3],
                        1e-5,
                        self._std_weight_vel * bbox[3]]
            self.filter.processNoiseCov = np.diag(np.square(proc_cov))
            # ~~ till here follows deepSORT
        else:
            # ~~~~ This is according to SORT
            self.filter.processNoiseCov = np.eye(2 * self.NDIM)
            # self.filter.processNoiseCov[2, 2] = 1e-2
            self.filter.processNoiseCov[self.NDIM:, self.NDIM:] *= 0.01
            self.filter.processNoiseCov[-2:, -2:] *= 0.01
            # ~~~~ Till here is according to SORT

        # Measurement noise covariance R
        if not self.cov_deepsort:
            # ~~~~ This is according to SORT
            self.filter.measurementNoiseCov = np.eye(self.NDIM)
            self.filter.measurementNoiseCov[2:, 2:] *= 10.0
            # ~~~~ Till here is according to SORT
        self.filter.statePost = np.r_[au.tlwh2xyrh(bbox), np.zeros(self.NDIM)]

    @property
    def pos(self):
        return au.xyrh2tlwh(self.filter.statePost[: self.NDIM])

    def predict(self):
        if self.cov_deepsort:
            # ~~ This follows deepSORT
            proc_cov = [self._std_weight_pos * self.filter.statePost[3],
                        self._std_weight_pos * self.filter.statePost[3],
                        1e-2,
                        self._std_weight_pos * self.filter.statePost[3],
                        self._std_weight_vel * self.filter.statePost[3],
                        self._std_weight_vel * self.filter.statePost[3],
                        1e-5,
                        self._std_weight_vel * self.filter.statePost[3]]
            self.filter.processNoiseCov = np.diag(np.square(proc_cov))
            # ~~ till here follows deepSORT
        if self.time_since_update > 0:
            self.hits = 0
        self.time_since_update += 1
        ret = self.filter.predict()
        return au.xyrh2tlwh(ret[:self.NDIM].squeeze())

    def update(self, detection):
        if self.cov_deepsort:
            # ~~ This follows deepSORT
            measure_cov = [self._std_weight_pos * self.filter.statePost[3],
                           self._std_weight_pos * self.filter.statePost[3],
                           1e-1,
                           self._std_weight_pos * self.filter.statePost[3]]
            self.filter.measurementNoiseCov = np.diag(np.square(measure_cov))
            # ~~ till here follows deepSORT
        pos = self.filter.correct(au.tlwh2xyrh(detection))
        self.time_since_update = 0
        self.hits += 1
        self.pos[:] = pos[:self.NDIM]
        return self.pos


class SORTracker(object):
    """SORT algorithm implementation

    NOTE: accepts bounding boxes in (x, y, w, h) format.
    """
    sigTracked = qc.pyqtSignal(dict, int)

    def __init__(self, metric=argos.constants.DistanceMetric.iou, min_dist=0.3,
                 max_age=1,
                 n_init=3, min_hits=3,
                 boxtype=argos.constants.OutlineStyle.bbox):
        super(SORTracker, self).__init__()
        self.n_init = n_init
        self.min_hits = min_hits
        self.boxtype = boxtype
        self.metric = metric
        if self.metric == argos.constants.DistanceMetric.iou:
            self.min_dist = 1 - min_dist
        else:
            self.min_dist = min_dist
        self.max_age = max_age
        self.trackers = {}
        self._next_id = 1
        self.frame_count = 0

    def reset(self):
        logging.debug('Resetting trackers.')
        self.trackers = {}
        self._next_id = 1
        self.frame_count = 0

    def setDistMetric(self, metric: argos.constants.DistanceMetric) -> None:
        if self.metric != metric:
            self.metric = metric
            if self.min_dist < 1:
                self.min_dist = 1 - self.min_dist

    def setMinDist(self, dist: float) -> None:
        if self.metric == argos.constants.DistanceMetric.iou:
            self.min_dist = 1 - dist
        else:
            self.min_dist = dist

    def update(self, bboxes):
        """Update the trackers with new bboxes.

        Parameters
        ----------
        bboxes: np.ndarray
            a 2D array with one row for each bbox in (x, y, w, h) format.

        Returns
        -------
        dict
            dict mapping object ids to tracked position
        """
        predicted_bboxes = {}
        for id_, tracker in self.trackers.items():
            prior = tracker.predict()
            if np.any(np.isnan(prior)) or np.any(
                    prior[:KalmanTracker.NDIM] < 0):
                logging.info(f'Found nan or negative in prior of {id_}')
                continue
            predicted_bboxes[id_] = prior[:KalmanTracker.NDIM]
        self.trackers = {id_: self.trackers[id_] for id_ in predicted_bboxes}
        for id_, bbox in predicted_bboxes.items():
            if np.any(bbox < 0):
                logging.debug(f'EEEE prediced bbox negative: {id_}: {bbox}')
        matched, new_unmatched, old_unmatched = match_bboxes(
            predicted_bboxes,
            bboxes[:, :KalmanTracker.NDIM],
            boxtype=self.boxtype,
            metric=self.metric,
            max_dist=self.min_dist)
        for track_id, bbox_id in matched.items():
            self.trackers[track_id].update(bboxes[bbox_id])
        for ii in new_unmatched:
            self._add_tracker(bboxes[ii, :KalmanTracker.NDIM])
        ret = {}
        for id_ in list(self.trackers.keys()):
            tracker = self.trackers[id_]
            if (tracker.time_since_update < 1) and \
                    (tracker.hits >= self.min_hits or
                     self.frame_count <= self.min_hits):
                ret[id_] = tracker.pos
            if tracker.time_since_update > self.max_age:
                self.trackers.pop(id_)
        return ret

    def _add_tracker(self, bbox):
        self.trackers[self._next_id] = KalmanTracker(bbox, self._next_id,
                                                     self.n_init,
                                                     self.max_age)
        self._next_id += 1





# -*- coding: utf-8 -*-
# Author: Subhasis Ray <ray dot subhasis at gmail dot com>
# Created: 2020-07-14 1:39 PM
"""
===============================================
Track objects in batch mode (non-interactively)
===============================================
Usage:
::
     python -m argos_track.batchtrack -i {input_file} -o {output_file}
     -c {config_file}

Try ``python -m argos_track.batchtrack -h`` for details of command-line
options.

This program allows non-interactive tracking of objects in a video.
When using classical segmentation this can speed things up by
utilizing multiple CPU cores.

It may be easier to use the interactive tracking :py:mod:`argos_track`
to play with the segmentation parameters to see what work best for
videos in a specific setting. The optimal setting can then be exported
to a configuration file which will then be passed with ``-c`` command
line option .

Examples
--------
Use YOLACT for segmentation and SORT for tracking:
::
    python -m argos_track.batchtrack -i video.avi -o video.h5 -m yolact \\
    --yconfig=config/yolact.yml -w config/weights.pth -s 0.1 -k 10 \\
    --overlap_thresh=0.3 --cuda \\
    --pmin=10 --pmax=500 --wmin=5 --wmax=100 --hmin=5 --hmax=100 \\
    -x 0.3 --min_hits=3 --max_age=20

The above command tells the ``batchtrack`` script to read the input
video ``video.avi`` and write the output to the file ``video.h5``. The
rest of the arguments:

- ``-m yolact`` tells it to use YOLACT as the segmentation method.

- ``--yconfig=config/yolact.yml``: Read YOLACT settings from the file
``config/yolact.yml``

- ``-w config/weights.pth``: Read YOLACT neural network weights from
the file ``config/weights.pth``.

- ``-s 0.1``: Include detections with score above 0.1

- ``-k 10``: Keep only the top 10 detections.

- ``--overlap_thresh=0.3``: At segmentation stage, merge detections
  whose bounding boxes overlap more than 0.3 of their total area.

- ``--cuda``: use GPU acceleration.

- ``--pmin=10``: Include objects at least 10 pixels in bounding box
  area.

- ``--pmax=500``: Include objects at most 500 pixels in bounding box
  area.

- ``--wmin=5``: Include objects at least 5 pixels wide.

- ``--wmax=100``: Include objects at most 100 pixels wide.

- ``--hmin=5``: Include objects at least 5 pixels long.

- ``--hmax=100``: Include objects at most 100 pixels long.

- ``-x 0.3``: In the tracking stage, if objects in two successive
  frames overlap more than 0.3 times their combined area, then
  consider them to be the same object.

- ``--min_hits=3``: An object must be detcted at least in 3
  consecutive frames to be included in the tracks.

- ``--max_age=20``: If an object cannot be matched to any detected
  object across 20 successive frames, then discard it (possibly it
  exited the view). [Remember that if you have a 30 frames per second
  video, 20 frames means 2/3 second in real time.]


All of this can be more easily set graphically in
:py:mod:`argos_track` and exported into a file, which can then be
passed with ``-c {config_file}``.

"""
import argparse
import os
from collections import namedtuple
import sys
import logging
from functools import partial, wraps
import numpy as np
import cv2
import yaml
import time
from datetime import timedelta
import pandas as pd
import concurrent.futures as cf
import torch
import torch.multiprocessing as mp
import torch.backends.cudnn as cudnn
from yolact import Yolact
from yolact.data import config as yconfig
# This is actually yolact.utils
from yolact.utils.augmentations import FastBaseTransform
from yolact.layers import output_utils as oututils
from argos_track.sortracker import SORTracker
from argos.constants import DistanceMetric, OutlineStyle
import argos.utility as ut
from argos.segment import (
    segment_by_contour_bbox,
    segment_by_dbscan,
    segment_by_watershed,
    extract_valid
)

# NOTE: defaults in namedtuple allowed only in Python3.7+
ThreshParam = namedtuple('ThreshParam', ('blur_width',
                                         'blur_sd',
                                         'invert',
                                         'method',
                                         'max_intensity',
                                         'baseline',
                                         'blocksize'),
                         defaults=(7, 1, True, 'gaussian', 255, 10, 41))

LimitParam = namedtuple('LimitParam', ('pmin',
                                       'pmax',
                                       'wmin',
                                       'wmax',
                                       'hmin',
                                       'hmax'),
                        defaults=(10, 500, 10, 50, 10, 200))

logging.basicConfig(stream=sys.stdout,
                    format='%(asctime)s '
                           'p=%(processName)s[%(process)d] '
                           't=%(threadName)s[%(thread)d] '
                           '%(filename)s#%(lineno)d:%(funcName)s: '
                           '%(message)s',
                    level=logging.INFO)

try:
    mp.set_start_method('spawn')
except RuntimeError:
    pass


def timed(func):
    @wraps(func)
    def _time_it(*args, **kwargs):
        tic = time.perf_counter_ns()
        try:
            return func(*args, **kwargs)
        finally:
            toc = time.perf_counter_ns()
            dt = (toc - tic) / 1e9
            dt = timedelta(seconds=dt)
            msg = f'{func}: Total execution time: {dt}'
            print(msg)
            logging.info(msg)

    return _time_it


# Global yolact network weights
ynet = None
config = None


def load_config(filename):
    """Load YOLACT configuration."""
    global config
    config = yconfig.cfg
    if filename == '':
        return
    print(f'Loading config from {filename}')
    with open(filename, 'r') as cfg_file:
        cfg = yaml.safe_load(cfg_file)
        for key, value in cfg.items():
            config.__setattr__(key, value)
        if 'mask_proto_debug' not in cfg:
            config.mask_proto_debug = False


def load_weights(filename, cuda):
    """Load YOLACT network weights"""
    global ynet
    if filename == '':
        raise ValueError('Empty filename for network weights')
    print('#### CUDA ENABLED', cuda)
    print(f'Loading weights from {filename}')
    tic = time.perf_counter_ns()
    with torch.no_grad():
        if cuda:
            cudnn.fastest = True
            torch.set_default_tensor_type('torch.cuda.FloatTensor')
        else:
            torch.set_default_tensor_type('torch.FloatTensor')
        # torch.set_default_tensor_type('torch.FloatTensor')
        ynet = Yolact()
        ynet.load_weights(filename, False)
        ynet.eval()
    toc = time.perf_counter_ns()
    logging.debug(f'Time to load weights: {1e-9 * (toc - tic)}')


@timed
def init_yolact(cfgfile, netfile, cuda):
    """Initialize YOLACT"""
    load_config(cfgfile)
    load_weights(netfile, cuda)


# This function should stay here for it uses the globals
# @timed
def segment_yolact(frame, score_threshold, top_k, overlap_thresh, cfgfile,
                   netfile, cuda):
    """Segment objects in frame using YOLACT.

    Parameters
    ----------
    frame: numpy.ndarray
        (WxHxC) integer array with the image content.
    score_threshold: float
        Minimum score to include object, should be in `(0, 1)`.
    top_k: int
        The number of segmented objects to keep.
    overlap_thresh: float
        Merge objects whose bounding boxes overlap (intersection over union)
        more than this amount.
    cfgfile: str
        Path to YOLACT configuration file.
    netfile: str
        Path to YOLACT network weights file.
    cuda: bool
        Whether to use CUDA.
    Returns
    -------
    numpy.ndarray
        An array of bounding boxes of detected objects in
        (xleft, ytop, width, height) format.
    """
    global ynet
    global config

    if ynet is None:
        init_yolact(cfgfile, netfile, cuda)
    # Partly follows yolact eval.py
    tic = time.perf_counter_ns()
    with torch.no_grad():
        if cuda:
            frame = torch.from_numpy(frame).cuda().float()
        else:
            frame = torch.from_numpy(frame).float()
        batch = FastBaseTransform()(frame.unsqueeze(0))
        preds = ynet(batch)
        h, w, _ = frame.shape
        config.rescore_bbox = True
        classes, scores, boxes, masks = oututils.postprocess(
            preds, w, h,
            visualize_lincomb=False,
            crop_masks=True,
            score_threshold=score_threshold)
        idx = scores.argsort(0, descending=True)[:top_k]
        # if self.config.eval_mask_branch:
        #     masks = masks[idx]
        classes, scores, boxes = [x[idx].cpu().numpy()
                                  for x in (classes, scores, boxes)]
        # This is probably not required, `postprocess` uses
        # `score_thresh` already
        # num_dets_to_consider = min(self.top_k, classes.shape[0])
        # for j in range(num_dets_to_consider):
        #     if scores[j] < self.score_threshold:
        #         num_dets_to_consider = j
        #         break
        # logging.debug('Bounding boxes: %r', boxes)
        # Convert from top-left bottom-right format to
        # top-left, width, height format
        if len(boxes) == 0:
            return np.empty(0)

        boxes[:, 2:] = boxes[:, 2:] - boxes[:, :2]
        boxes = np.asanyarray(np.rint(boxes), dtype=np.int_)
        if overlap_thresh < 1:
            dist_matrix = ut.pairwise_distance(new_bboxes=boxes, bboxes=boxes,
                                               boxtype=OutlineStyle.bbox,
                                               metric=DistanceMetric.iou)
            bad_boxes = []
            for ii in range(dist_matrix.shape[0] - 1):
                for jj in range(ii + 1, dist_matrix.shape[1]):
                    if dist_matrix[ii, jj] < 1 - overlap_thresh:
                        bad_boxes.append(jj)
            boxes = np.array([boxes[ii] for ii in range(boxes.shape[0]) if
                              ii not in bad_boxes], dtype=np.int_)
        toc = time.perf_counter_ns()
        logging.debug('Time to process single image: %f s',
                      1e-9 * (toc - tic))
        return boxes


def read_frame(video):
    """Read a frame from `video` and return the frame number and the image data"""
    if (video is None) or not video.isOpened():
        return (-1, None)
    frame_no = int(video.get(cv2.CAP_PROP_POS_FRAMES))
    ret, frame = video.read()
    return (frame_no, frame)


def threshold(frame: np.ndarray, params: ThreshParam):
    """Adaptive threshold image after converting to grayscale and blurring.

    Parameters
    ----------
    frame: numpy.ndarray
        Image data.
    params: ThreshParam
        Parameters for thresholding.
    Returns
    -------
    np.ndarray
        Binary image from thresholding.
    """
    gray = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
    gray = cv2.GaussianBlur(gray, ksize=(params.blur_width, params.blur_width),
                            sigmaX=params.blur_sd)
    if params.invert:
        thresh_type = cv2.THRESH_BINARY_INV
    else:
        thresh_type = cv2.THRESH_BINARY
    if params.method == 'gaussian':
        thresh_method = cv2.ADAPTIVE_THRESH_GAUSSIAN_C
    elif params.method == 'mean':
        thresh_method = cv2.ADAPTIVE_THRESH_MEAN_C
    else:
        raise ValueError(f'Invalid thresholding method {params.method}')
    binary = cv2.adaptiveThreshold(gray,
                                   maxValue=params.max_intensity,
                                   adaptiveMethod=thresh_method,
                                   thresholdType=thresh_type,
                                   blockSize=params.blocksize,
                                   C=params.baseline)
    return binary


def bbox_func(points_list):
    """Calculate list of bounding boxes of point arrays"""
    ret = np.array([cv2.boundingRect(points) for points in points_list])
    return ret


def create_seg_func_list(args):
    """Create a sequence of functions to be applied on image for
    segmentation using classical methods.

    Classical segmentation involves blurring and thresholding followed
    by additional procedures depending on the specific method.

    Parameters
    ----------
    args: dict
        Argument dictionary.
    Returns
    -------
    list
        List of functions to be applied in sequence on the input image.
    """
    thresh_params = ThreshParam(blur_width=args.blur_width,
                                blur_sd=args.blur_sd,
                                invert=args.thresh_invert,
                                method=args.thresh_method,
                                max_intensity=args.thresh_max,
                                baseline=args.thresh_baseline,
                                blocksize=args.thresh_blocksize)
    thresh_func = partial(threshold, params=thresh_params)

    seg_method = args.seg_method
    if seg_method == 'threshold':
        seg_func = segment_by_contour_bbox
    elif seg_method == 'watershed':
        seg_func = partial(segment_by_watershed, dist_thresh=args.dist_thresh)
    elif seg_method == 'dbscan':
        seg_func = partial(segment_by_dbscan, eps=args.eps,
                           min_samples=args.min_samples)
    else:
        raise ValueError(f'Unknown segmentation method: {seg_method}')

    limit_func = partial(extract_valid,
                         pmin=args.pmin,
                         pmax=args.pmax,
                         wmin=args.wmin,
                         wmax=args.wmax,
                         hmin=args.hmin,
                         hmax=args.hmax)
    return [thresh_func, seg_func, limit_func, bbox_func]


# @timed
def run_fn_seq(fn_args):
    """Run frame through a function pipeline.

    Parameters
    ----------
    fn_args: tuple
        tuple of the form ((f0, a0), (f1, a1), ...) where `f0`, `f1`, etc. are
         functions and `a0`, `a1`, etc. are the arguments (tuple) of the
         corresponding function.
    Could not use `reduce` as the number of arguments varies depending  on the
    function.
    """
    result = None
    for fn, args in fn_args:
        if result is None:
            result = fn(*args)
        else:
            result = fn(result, *args)
    return result


@timed
def batch_segment(args):
    """Segment frames in parallel and save the bboxes of segmented objects in
    an HDF file for later tracking"""
    if 'SLURM_JOB_ID' in os.environ:  # this is a slurm job, use slurm env var or safe number of 2
        cpu_count = int(os.environ.get('SLURM_CPUS_PER_TASK', '2'))
    else:
         cpu_count = mp.cpu_count()
    max_workers = args.max_proc if args.max_proc > 0 else cpu_count
    print('BBBBBBB cuda', args.cuda)
    if args.seg_method == 'yolact':
        seg_fn = partial(segment_yolact, score_threshold=args.score_thresh,
                         top_k=args.top_k,
                         overlap_thresh=args.overlap_thresh,
                         cfgfile=args.yconfig, netfile=args.weight,
                         cuda=args.cuda)
        max_workers = max(1, min(cpu_count, torch.cuda.device_count()))
    else:
        thresh_fn, seg_fn, limit_fn, bbox_fn = create_seg_func_list(args)
    cv2.setNumThreads(cpu_count)
    torch.set_num_threads(cpu_count)
    print('Workers:', max_workers, 'CPUs:', cpu_count)
    logging.info(f'Running segmentation with {max_workers} worker processes on system with {cpu_count} cpus.')
    video = cv2.VideoCapture(args.infile)
    if not video.isOpened():
        raise IOError('Could not open video')
    data = []
    with cf.ProcessPoolExecutor(max_workers=max_workers) as executor:
        while video.isOpened():
            frames = []
            for ii in range(max_workers):
                frame_no, frame = read_frame(video)
                if frame_no % 100 == 0:
                    logging.debug(f'Read till {frame_no}')
                if frame is None:
                    video.release()
                    break
                frames.append((frame_no, frame))
            futures = {}
            for frame_no, frame in frames:
                if args.seg_method == 'yolact':
                    fut = executor.submit(seg_fn, frame)
                else:
                    seg_arg = (frame,) if args.seg_method == 'watershed' else ()
                    fut = executor.submit(run_fn_seq, ((thresh_fn, (frame,)),
                                                       (seg_fn, seg_arg),
                                                       (limit_fn, ()),
                                                       (bbox_fn, ())))
                futures[fut] = frame_no
            while futures:
                done, _ = cf.wait(futures, return_when=cf.FIRST_COMPLETED)
                for fut in done:
                    frame_no = futures.pop(fut)
                    result = fut.result()
                    for ii, bbox in enumerate(result):
                        if len(bbox) > 0:
                            data.append({'frame': frame_no,
                                         'x': bbox[0],
                                         'y': bbox[1],
                                         'w': bbox[2],
                                         'h': bbox[3]})
    if len(data) == 0:
        raise RuntimeError('Data list empty')
    data = pd.DataFrame(data)
    data.sort_values(by='frame', inplace=True)
    data.to_hdf(args.outfile, 'segmented')
    logging.info(f'Data saved in {args.outfile} under /segmented')


@timed
def batch_track(args):
    """outfile should have a `/segmented` table containing the segmentation data.

    saves the result in same file under `/tracked`
    """
    segments = pd.read_hdf(args.outfile, 'segmented')
    results = []
    if args.sort_metric == 'iou':
        metric = DistanceMetric.iou
    else:
        metric = DistanceMetric.euclidean
    tracker = SORTracker(metric=metric,
                         min_dist=args.min_dist,
                         max_age=args.max_age,
                         n_init=args.min_hits,
                         min_hits=args.min_hits)
    for frame, fgrp in segments.groupby('frame'):
        if len(fgrp) == 0:
            continue
        tracked = tracker.update(fgrp[['x', 'y', 'w', 'h']].astype(np.int_).values)
        for tid, bbox in tracked.items():
            results.append({'frame': frame,
                            'trackid': tid,
                            'x': bbox[0],
                            'y': bbox[1],
                            'w': bbox[2],
                            'h': bbox[3]})

        if frame % 100 == 0:
            logging.info(f'Processed till {frame}')
    results = pd.DataFrame(results)
    results.sort_values(by='frame', inplace=True)
    results.to_hdf(args.outfile, 'tracked')
    logging.info(f'Tracking data saved in {args.outfile} under /tracked.')


def make_parser():
    parser = argparse.ArgumentParser('Track objects in video in batch mode')
    parser.add_argument('-i', '--infile', type=str, help='input file')
    parser.add_argument('-o', '--outfile', type=str,
                        help='output file. Create an HDF file with segmentation'
                             ' in `segmented` and tracking data in the table'
                             ' `tracked`')
    parser.add_argument('-c', '--config', type=str,
                        help='configuration file to use for rest of the'
                             ' arguments')
    parser.add_argument('-p', '--max_proc', type=int, default=-1,
                        help='number of parallel processes')
    parser.add_argument('-m', '--seg_method', type=str, default='yolact',
                        help='method for segmentation'
                             ' (yolact/threshold/watershed/dbscan)')
    yolact_grp = parser.add_argument_group('YOLACT',
                                           'Parameters for YOLACT segmentation')
    yolact_grp.add_argument('--yconfig', type=str,
                            help='YOLACT configuration file')
    yolact_grp.add_argument('-w', '--weight', type=str,
                            help='YOLACT trained weights file')
    yolact_grp.add_argument('-s', '--score_thresh', type=float, default=0.3,
                            help='score threshold for accepting a detected object')
    yolact_grp.add_argument('-k', '--top_k', type=int, default=30,
                            help='maximum number of objects above score'
                                 ' threshold to keep')
    yolact_grp.add_argument('--overlap_thresh', type=float, default=1.0,
                            help='Bboxes with IoU overlap higher than this are'
                                 ' merged')
    yolact_grp.add_argument('--cuda', action='store_true',
                            help='If specified, use CUDA')
    thresh_grp = parser.add_argument_group(
        'Threshold',
        'Parameters for thresholding'
    )
    thresh_grp.add_argument('--thresh_method', type=str, default='gaussian',
                            help='Method for adaptive thresholding'
                                 ' (gaussian/mean)')
    thresh_grp.add_argument('--thresh_max', type=int, default=255,
                            help='Maximum intensity for thresholding')
    thresh_grp.add_argument('--thresh_baseline', type=int, default=10,
                            help='baseline intensity for thresholding')
    thresh_grp.add_argument('--blur_width', type=int, default=7,
                            help='blur width before thresholding.'
                                 ' Must be odd number.')
    thresh_grp.add_argument('--blur_sd', type=float, default=1.0,
                            help='SD for Gaussian blur before thresholding')
    thresh_grp.add_argument('--thresh_blocksize', type=int, default=41,
                            help='block size for adaptive thresholding.'
                                 ' Must be odd number')
    thresh_grp.add_argument('--thresh_invert', type=bool, default=True,
                            help='Inverted thresholding')
    watershed_grp = parser.add_argument_group(
        'Watershed',
        'Parameter for segmentation using watershed algorithm')
    watershed_grp.add_argument('-d', '--dist_thresh', type=float, default=3.0,
                               help='minimum distance of pixels from detected '
                                    'boundary to consider them core points')
    dbscan_grp = parser.add_argument_group(
        'DBSCAN',
        'Parameters for segmentation by clustering pixels with'
        ' DBSCAN algorithm')
    dbscan_grp.add_argument('-e', '--eps', type=float, default=5.0,
                            help='epsilon parameter for DBSCAN')
    dbscan_grp.add_argument('--min_samples', type=int, default=10,
                            help='minimum number of pixels in each cluster'
                                 ' for DBSCAN')
    limits_grp = parser.add_argument_group(
        'Limits',
        'Parameters to set limits on detected object size.')
    limits_grp.add_argument('--pmin', type=int, default=10,
                            help='Minimum number of pixels')
    limits_grp.add_argument('--pmax', type=int, default=500,
                            help='Maximum number of pixels')
    limits_grp.add_argument('--hmin', type=int, default=10,
                            help='Minimum height (longer side) of bounding box'
                                 ' in pixels')
    limits_grp.add_argument('--hmax', type=int, default=200,
                            help='Maximum height (longer side) of bounding box'
                                 ' in pixels')
    limits_grp.add_argument('--wmin', type=int, default=10,
                            help='Minimum width (shorter side) of bounding box'
                                 ' in pixels')
    limits_grp.add_argument('--wmax', type=int, default=100,
                            help='Maximum width (shorter side) of bounding box'
                                 ' in pixels')
    track_grp = parser.add_argument_group(
        'Tracker',
        'Parameters for SORT tracker')
    track_grp.add_argument('--sort_metric', type=str, default='iou',
                           help='Metric for measuring closeness.'
                           ' iou or euclidean')
    track_grp.add_argument('-x', '--min_dist', type=float, default=0.3,
                           help='Minimum distance between bounding boxes.'
                           ' If iou, this their intersection as a'
                           ' fraction of their combined area.'
                           ' If euclidean, number of pixels.')
    track_grp.add_argument('--min_hits', type=int, default=3,
                           help='Minimum number of hits to accept a track')
    track_grp.add_argument('--max_age', type=int, default=50,
                           help='Maximum number of misses to exclude a track')
    parser.add_argument('--debug', action='store_true', help='Print debug info')
    return parser


if __name__ == '__main__':
    logging.getLogger().setLevel(logging.INFO)
    # 2 proc 40 / 124 fps
    # 5 proc 25 / 50 fps
    parser = make_parser()
    args = parser.parse_args()
    print('ARGS:')
    print(args)
    if args.config is not None:
        with open(args.config, 'r') as cfg_file:
            config = yaml.safe_load(cfg_file)
            for key, value in config.items():
                vars(args)[key] = value
        print('ARGS after update from config:')
        print(args)
    if args.debug:
        logging.getLogger().setLevel(logging.DEBUG)
    else:
        logging.getLogger().setLevel(logging.INFO)
    batch_segment(args)
    batch_track(args)
    print('Finished tracking')



# -*- coding: utf-8 -*-
# Author: Subhasis Ray <ray dot subhasis at gmail dot com>
# Created: 2020-11-28 11:45 PM
import logging
import time
import numpy as np
from collections import OrderedDict
from PyQt5 import QtCore as qc, QtWidgets as qw

import argos.constants
from argos_track.sortracker import SORTracker, settings

# Ordered dict to maintain sequence
distance_metric = OrderedDict(
    (('Intersection over Union', argos.constants.DistanceMetric.iou),
     ('Euclidean', argos.constants.DistanceMetric.euclidean))
)


class QSORTracker(qc.QObject):
    """Qt Wrapper class over SORTracker to allow signal/slot communication"""
    sigTracked = qc.pyqtSignal(dict, int)

    def __init__(self, *args, **kwargs):
        super(QSORTracker, self).__init__()
        self.tracker = SORTracker(*args, **kwargs)
        self._mutex = qc.QMutex()

    @qc.pyqtSlot()
    def reset(self):
        _ = qc.QMutexLocker(self._mutex)
        self.tracker.reset()

    @qc.pyqtSlot(argos.constants.DistanceMetric)
    def setDistMetric(self, metric):
        self.tracker.setDistMetric(metric)

    @qc.pyqtSlot(float)
    def setMinDist(self, dist: float) -> None:
        _ = qc.QMutexLocker(self._mutex)
        self.tracker.setMinDist(dist)
        metric = settings.value('sortracker/metric', 'iou', type=str)
        if metric == 'iou':
            settings.setValue('sortracker/iou_mindist', dist)
        else:
            settings.setValue('sortracker/euclidean_mindist', dist)

    @qc.pyqtSlot(int)
    def setMaxAge(self, max_age: int) -> None:
        """Set the maximum misses before discarding a track"""
        _ = qc.QMutexLocker(self._mutex)
        settings.setValue('sortracker/max_age', max_age)
        self.tracker.max_age = max_age

    @qc.pyqtSlot(int)
    def setMinHits(self, count: int) -> None:
        """Number of times a track should match prediction before it is
        confirmed"""
        _ = qc.QMutexLocker(self._mutex)
        settings.setValue('sortracker/min_hits', count)
        self.tracker.n_init = count

    @qc.pyqtSlot(dict, int)
    def track(self, bboxes: np.ndarray, pos: int) -> None:
        _ts = time.perf_counter()
        logging.debug(f'Received from {self.sender()} bboxes: {bboxes}')
        _ = qc.QMutexLocker(self._mutex)
        if len(bboxes) == 0:
            ret = {}
        else:
            ret = self.tracker.update(bboxes)
        logging.debug(f'SORTracker: frame {pos}, Rectangles: \n{ret}')
        self.sigTracked.emit(ret, pos)
        _dt = time.perf_counter() - _ts
        logging.debug(f'{__name__}.{self.__class__.__name__}.track: Runtime: {_dt}s')

class SORTWidget(qw.QWidget):
    sigTrack = qc.pyqtSignal(np.ndarray, int)
    sigTracked = qc.pyqtSignal(dict, int)
    sigQuit = qc.pyqtSignal()
    sigReset = qc.pyqtSignal()
    sigDistMetric = qc.pyqtSignal(argos.constants.DistanceMetric)

    def __init__(self, *args, **kwargs):
        super(SORTWidget, self).__init__(*args, **kwargs)
        self._max_age_label = qw.QLabel('Maximum age')
        self._max_age_label.setToolTip(
            'Maximum number of misses before a track is removed')
        self._max_age_spin = qw.QSpinBox()
        self._max_age_spin.setRange(1, 100)
        value = settings.value('sortracker/max_age', 10, type=int)
        self._max_age_spin.setValue(value)
        self._max_age_spin.setToolTip(self._max_age_label.toolTip())
        self._conf_age_label = qw.QLabel('Minimum hits')
        self._conf_age_label.setToolTip(
            'Minimum number of hits before a track is confirmed')
        self._conf_age_spin = qw.QSpinBox()
        self._conf_age_spin.setRange(1, 100)
        value = settings.value('sortracker/min_hits', 3, type=int)
        self._conf_age_spin.setValue(value)
        self._conf_age_spin.setToolTip(self._conf_age_label.toolTip())

        self._dist_metric_label = qw.QLabel('Distance metric')
        self._dist_metric_combo = qw.QComboBox()
        self._dist_metric_combo.addItems(list(distance_metric.keys()))
        metric = settings.value('sortracker/metric', 'iou', type=str)
        self._min_dist_spin = qw.QDoubleSpinBox()
        try:
            self._min_dist_spin.setStepType(
                qw.QDoubleSpinBox.AdaptiveDecimalStepType)
        except AttributeError:
            pass     # older versions of Qt don't support this
        if metric == 'iou':
            self._min_dist_label = qw.QLabel('Minimum overlap')
            self._min_dist_spin.setRange(0.1, 1.0)
            self._min_dist_spin.setSingleStep(0.05)
            value = settings.value('sortracker/iou_mindist', 0.3, type=float)
            self._min_dist_spin.setValue(value)
            self._min_dist_spin.setToolTip(
                'Minimum overlap between bounding boxes '
                'to consider them same object.')
        else:
            self._min_dist_label = qw.QLabel('Maximum distance')
            self._min_dist_spin.setRange(1, 1000)
            self._min_dist_spin.setSingleStep(5)
            self._min_dist_spin.setValue(
                settings.value('sortracker/euclidean_mindist', 5, type=float))
            self._min_dist_spin.setToolTip(
                'Maximum distance between bounding boxes '
                'to consider them same object.')

        self._disable_check = qw.QCheckBox('Disable tracking')
        self._disable_check.setToolTip('Just show the identified objects. Can '
                                       'be useful for troubleshooting.')
        layout = qw.QFormLayout()
        self.setLayout(layout)
        layout.addRow(self._dist_metric_label, self._dist_metric_combo)
        layout.addRow(self._min_dist_label, self._min_dist_spin)
        layout.addRow(self._conf_age_label, self._conf_age_spin)
        layout.addRow(self._max_age_label, self._max_age_spin)
        layout.addWidget(self._disable_check)
        self.qtracker = QSORTracker(metric=argos.constants.DistanceMetric.iou,
                                    min_dist=self._min_dist_spin.value(),
                                    max_age=self._max_age_spin.value(),
                                    n_init=self._conf_age_spin.value())
        self.thread = qc.QThread()
        self.qtracker.moveToThread(self.thread)
        self._max_age_spin.valueChanged.connect(self.qtracker.setMaxAge)
        self._dist_metric_combo.currentTextChanged.connect(self.setDistMetric)
        self._min_dist_spin.valueChanged.connect(self.qtracker.setMinDist)
        self._conf_age_spin.valueChanged.connect(self.qtracker.setMinHits)
        self._disable_check.stateChanged.connect(self.disable)
        self.sigDistMetric.connect(self.qtracker.setDistMetric)
        self.sigTrack.connect(self.qtracker.track)
        self.qtracker.sigTracked.connect(self.sigTracked)
        self.sigReset.connect(self.qtracker.reset)
        self.sigQuit.connect(self.thread.quit)
        self.thread.finished.connect(self.thread.deleteLater)
        self.thread.start()

    @qc.pyqtSlot(str)
    def setDistMetric(self, text):
        metric = distance_metric[text]
        if metric == argos.constants.DistanceMetric.iou:
            # print('SORT Metric set to IoU')
            settings.setValue('sortracker/metric', 'iou')
            self._min_dist_label.setText('Minimum overlap')
            self._min_dist_spin.setRange(0.01, 1)
            self._min_dist_spin.setValue(
                settings.value('sortracker/iou_mindist', 0.3, type=float))
            self._min_dist_spin.setToolTip(
                'Minimum overlap between bounding boxes '
                'to consider them same object.')
        else:
            settings.setValue('sortracker/metric', 'euclidean')
            self._min_dist_label.setText('Maximum distance')
            self._min_dist_spin.setRange(1, 1000)
            self._min_dist_spin.setValue(
                settings.value('sortracker/euclidean_mindist', 5, type=float))
            self._min_dist_spin.setToolTip(
                'Maximum distance between bounding boxes '
                'to consider them same object.')
        self.sigDistMetric.emit(metric)

    @qc.pyqtSlot(int)
    def disable(self, state):
        self.sigTrack.disconnect()
        if state:
            self.sigTrack.connect(self.sendDummySigTracked)
        else:
            self.sigTrack.connect(self.qtracker.track)

    @qc.pyqtSlot(np.ndarray, int)
    def sendDummySigTracked(self, bboxes: np.ndarray, pos: int) -> None:
        ret = {ii + 1: bboxes[ii] for ii in range(bboxes.shape[0])}
        self.sigTracked.emit(ret, pos)

    @qc.pyqtSlot(np.ndarray, int)
    def track(self, bboxes: np.ndarray, pos: int) -> None:
        """Just to intercept signal source for debugging"""
        logging.debug(
            f'Received frame {pos} from {self.sender()} bboxes: {bboxes}')
        self.sigTrack.emit(bboxes, pos)


def test():
    import sys
    app = qw.QApplication(sys.argv)
    win = SORTWidget()
    win.setMinimumSize(800, 600)
    # win.setWindowTitle('Argos - track animals in video')
    # win.showMaximized()
    # app.aboutToQuit.connect(win.cleanup)
    win.show()
    sys.exit(app.exec_())


if __name__ == '__main__':
    test()



# -*- coding: utf-8 -*-
# Author: Subhasis Ray <ray dot subhasis at gmail dot com>
# Created: 2020-06-04 4:13 PM
"""How to build distribution:

- Build wheel and tar (update MICRO-version number for test - once used, a filename cannot be reused)

      python -m build

- Check dist folder for old archives - delete them, rename files if required. Then upoload

      python -m twine upload --repository testpypi dist/*


- Install from TestPyPI

      pip install --index-url https://test.pypi.org/simple/ --no-deps argos-tracker --extra-index-url https://pypi.org/simple

This requires the argos-toolkit package.

``argos-tracker`` is PyPI name, the installed module is named
``argos_track``.
"""

from setuptools import setup, find_packages, Extension
import numpy


with open('README.md', 'r') as fh:
    long_description = fh.read()



setup(
    name='argos_tracker',
    version='0.1.1',
    author='Subhasis Ray',
    author_email='ray.subhasis@gmail.com',
    description='Software utility for tracking multiple objects (animals) in a video.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/subhacom/argos_tracker',
    project_urls={
    'Documentation': 'https://argos.readthedocs.io',
    'Source': 'https://github.com/subhacom/argos_tracker',
    'Tracker': 'https://github.com/subhacom/argos/issues',
    },
    packages=find_packages(),
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: GNU General Public License (GPL)',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: MacOS',
        'Operating System :: POSIX :: Linux',
        'Topic :: Scientific/Engineering',
        'Topic :: Utilities',
    ],
    python_requires='>=3.6',
    install_requires=[
        'torch',
        'torchvision',
        'numpy',
        'scipy',
        'scikit-learn',
        'pandas',
        'tables',
        'sortedcontainers',
        'pyqt5',
        'opencv-contrib-python',
        'pyyaml',
        'matplotlib',
        'argos_toolkit'
    ],
)



# -*- coding: utf-8 -*-
# Author: Subhasis Ray <ray dot subhasis at gmail dot com>
# Created: 2020-05-28 3:01 PM

import sys
import logging
import yaml
from PyQt5 import (
    QtWidgets as qw,
    QtCore as qc,
    QtGui as qg
)

import argos.utility as util
from argos.vwidget import VideoWidget
from argos.yolactwidget import YolactWidget
from argos_track.sortrackerwidget import SORTWidget
from argos.segwidget import SegWidget
from argos.csrtracker import CSRTWidget
from argos.limitswidget import LimitsWidget

# Set up logging for multithreading/multiprocessing
settings = util.init()


class ArgosTracker(qw.QMainWindow):
    """"Main user interface"""

    sigQuit = qc.pyqtSignal()

    def __init__(self, *args, **kwargs):
        super(ArgosTracker, self).__init__(*args, **kwargs)
        self._video_widget = VideoWidget()
        self.setCentralWidget(self._video_widget)
        self._yolact_widget = YolactWidget()
        self._seg_widget = SegWidget()
        self._seg_widget.fixBboxOutline()

        self._lim_widget = LimitsWidget()

        self._sort_widget = SORTWidget()
        self._csrt_widget = CSRTWidget()

        self._yolact_dock = qw.QDockWidget('Yolact settings')
        self._yolact_dock.setAllowedAreas(qc.Qt.LeftDockWidgetArea |
                                          qc.Qt.RightDockWidgetArea)
        self.addDockWidget(qc.Qt.RightDockWidgetArea, self._yolact_dock)
        self._yolact_dock.setWidget(self._yolact_widget)

        self._seg_dock = qw.QDockWidget('Segmentation settings')
        self._seg_dock.setAllowedAreas(qc.Qt.LeftDockWidgetArea |
                                       qc.Qt.RightDockWidgetArea)
        self.addDockWidget(qc.Qt.RightDockWidgetArea, self._seg_dock)
        self._seg_scroll = qw.QScrollArea()
        self._seg_scroll.setWidgetResizable(True)
        self._seg_scroll.setWidget(self._seg_widget)
        self._seg_dock.setWidget(self._seg_scroll)

        self._lim_dock = qw.QDockWidget('Size limits')
        self._lim_dock.setAllowedAreas(qc.Qt.LeftDockWidgetArea |
                                       qc.Qt.RightDockWidgetArea)
        self.addDockWidget(qc.Qt.RightDockWidgetArea, self._lim_dock)
        self._lim_dock.setWidget(self._lim_widget)

        self._sort_dock = qw.QDockWidget('SORTracker settings')
        self._sort_dock.setAllowedAreas(qc.Qt.LeftDockWidgetArea |
                                        qc.Qt.RightDockWidgetArea)
        self.addDockWidget(qc.Qt.RightDockWidgetArea, self._sort_dock)
        self._sort_scroll = qw.QScrollArea()
        self._sort_scroll.setWidgetResizable(True)
        self._sort_scroll.setWidget(self._sort_widget)
        self._sort_dock.setWidget(self._sort_scroll)

        self._csrt_dock = qw.QDockWidget('CSRTracker settings')
        self._csrt_dock.setAllowedAreas(qc.Qt.LeftDockWidgetArea |
                                        qc.Qt.RightDockWidgetArea)
        self._csrt_scroll = qw.QScrollArea()
        self._csrt_scroll.setWidgetResizable(True)
        self._csrt_scroll.setWidget(self._csrt_widget)
        self._csrt_dock.setWidget(self._csrt_scroll)
        self.addDockWidget(qc.Qt.RightDockWidgetArea, self._csrt_dock)
        self._yolact_action = qw.QAction('Use Yolact segmentation')
        self._seg_action = qw.QAction('Use classical segmentation')
        self._seg_grp = qw.QActionGroup(self)
        self._seg_grp.addAction(self._yolact_action)
        self._seg_grp.addAction(self._seg_action)
        self._seg_grp.setExclusive(True)
        self._yolact_action.setCheckable(True)
        self._seg_action.setCheckable(True)
        self._yolact_action.setChecked(True)
        self._seg_dock.hide()
        self._sort_action = qw.QAction('Use SORT for tracking')
        self._sort_action.setCheckable(True)
        self._csrt_action = qw.QAction('Use CSRT for tracking')
        self._csrt_action.setCheckable(True)
        self._track_grp = qw.QActionGroup(self)
        self._track_grp.addAction(self._sort_action)
        self._track_grp.addAction(self._csrt_action)
        self._sort_action.setChecked(True)
        self._debug_action = qw.QAction('Debug')
        self._debug_action.setCheckable(True)
        debug_level = settings.value('track/debug', logging.INFO, type=int)
        self._debug_action.setChecked(debug_level == logging.DEBUG)
        logging.getLogger().setLevel(debug_level)
        self._debug_action.triggered.connect(self.setDebug)
        self._clear_settings_action = qw.QAction('Reset to default settings')
        self._clear_settings_action.triggered.connect(self.clearSettings)
        self._csrt_dock.hide()
        self._menubar = self.menuBar()
        self._file_menu = self._menubar.addMenu('&File')
        self._file_menu.addAction(self._video_widget.openAction)
        self._file_menu.addAction(self._video_widget.openCamAction)
        self.saveConfigAction = qw.QAction('Save configuration in file')
        self.saveConfigAction.triggered.connect(self.saveConfig)
        self._file_menu.addAction(self.saveConfigAction)
        self._seg_menu = self._menubar.addMenu('&Segmentation method')
        self._seg_menu.addActions(self._seg_grp.actions())
        self._track_menu = self._menubar.addMenu('&Tracking method')
        self._track_menu.addActions(self._track_grp.actions())
        self._view_menu = self._menubar.addMenu('View')
        self._view_menu.addActions([self._video_widget.zoomInAction,
                                    self._video_widget.zoomOutAction,
                                    self._video_widget.resetArenaAction,
                                    self._video_widget.showGrayscaleAction,
                                    self._video_widget.setColorAction,
                                    self._video_widget.autoColorAction,
                                    self._video_widget.colormapAction,
                                    self._video_widget.setLabelInsideAction,
                                    self._video_widget.fontSizeAction,
                                    self._video_widget.relativeFontSizeAction,
                                    self._video_widget.lineWidthAction,
                                    self._video_widget.infoAction])
        self._play_menu = self.menuBar().addMenu('&Play')
        self._play_menu.addActions((self._video_widget.playAction,
                                    self._video_widget.refreshAction))
        self._advanced_menu = self.menuBar().addMenu('Advanced')
        # self._advanced_menu.addAction(self._video_widget.arenaSelectAction)
        self._advanced_menu.addAction(self._video_widget.resetArenaAction)
        self._advanced_menu.addAction(self._debug_action)
        self._advanced_menu.addAction(self._clear_settings_action)

        self.makeShortcuts()
        ##########################
        # Connections
        self._video_widget.sigVideoFile.connect(self.updateTitle)
        self._video_widget.sigSetFrame.connect(self._yolact_widget.process)
        self._video_widget.sigArena.connect(self._lim_widget.setRoi)
        self._video_widget.sigReset.connect(self._lim_widget.resetRoi)
        self._video_widget.sigArena.connect(self._seg_widget.setRoi)
        self._video_widget.sigReset.connect(self._seg_widget.resetRoi)
        self._video_widget.resetArenaAction.triggered.connect(
            self._seg_widget.resetRoi)
        self._yolact_widget.sigProcessed.connect(self._lim_widget.process)
        self._lim_widget.sigProcessed.connect(self._sort_widget.track)
        self._yolact_widget.sigProcessed.connect(
            self._video_widget.sigSetBboxes)
        self._lim_widget.sigWmin.connect(self._seg_widget.setWmin)
        self._lim_widget.sigWmax.connect(self._seg_widget.setWmax)
        self._lim_widget.sigHmin.connect(self._seg_widget.setHmin)
        self._lim_widget.sigHmax.connect(self._seg_widget.setHmax)
        self._seg_widget.sigProcessed.connect(self._sort_widget.sigTrack)
        self._seg_widget.sigProcessed.connect(self._video_widget.sigSetBboxes)
        # self._seg_widget.sigSegPolygons.connect(self._video_widget.sigSetSegmented)
        self._sort_widget.sigTracked.connect(self._video_widget.setTracked)
        self._csrt_widget.sigTracked.connect(self._video_widget.setTracked)
        self._video_widget.openAction.triggered.connect(
            self._sort_widget.sigReset)
        self._video_widget.openAction.triggered.connect(
            self._csrt_widget.sigReset)
        self._video_widget.sigReset.connect(self._sort_widget.sigReset)
        self._video_widget.sigReset.connect(self._csrt_widget.sigReset)
        self._video_widget.sigStatusMsg.connect(self.statusMsgSlot)
        self._seg_grp.triggered.connect(self.switchSegmentation)
        self._track_grp.triggered.connect(self.switchTracking)
        self.sigQuit.connect(self._video_widget.sigQuit)
        self.sigQuit.connect(self._yolact_widget.sigQuit)
        self.sigQuit.connect(self._seg_widget.sigQuit)
        self.sigQuit.connect(self._lim_widget.sigQuit)
        self.sigQuit.connect(self._sort_widget.sigQuit)
        self.sigQuit.connect(self._csrt_widget.sigQuit)

    def makeShortcuts(self):
        self.sc_play = qw.QShortcut(qg.QKeySequence(qc.Qt.Key_Space), self)
        self.sc_play.activated.connect(self._video_widget.playAction.trigger)
        self.sc_refresh = qw.QShortcut(qg.QKeySequence('R'), self)
        self.sc_refresh.activated.connect(self._video_widget.refreshAction.trigger)
        self.sc_open = qw.QShortcut(qg.QKeySequence('Ctrl+O'), self)
        self.sc_open.activated.connect(self._video_widget.openVideo)

    @qc.pyqtSlot(bool)
    def setDebug(self, state):
        level = logging.DEBUG if state else logging.INFO
        settings.setValue('track/debug', level)
        logging.getLogger().setLevel(level)

    def cleanup(self):
        self.sigQuit.emit()
        settings.sync()
        logging.debug('Saved settings')

    @qc.pyqtSlot()
    def clearSettings(self):
        button = qw.QMessageBox.question(
            self,
            'Reset settings to default',
            'Are you sure to clear all saved settings?')
        if button == qw.QMessageBox.NoButton:
            return
        settings.clear()
        settings.sync()

    @qc.pyqtSlot(qw.QAction)
    def switchSegmentation(self, action):
        """Switch segmentation widget between yolact and classical"""
        self._video_widget.pauseVideo()
        if action == self._yolact_action:
            util.reconnect(self._video_widget.sigSetFrame,
                           newhandler=self._yolact_widget.process,
                           oldhandler=self._seg_widget.sigProcess)
            self._yolact_dock.setVisible(True)
            self._seg_dock.setVisible(False)
        else:  # classical segmentation, self._seg_action
            util.reconnect(self._video_widget.sigSetFrame,
                           newhandler=self._seg_widget.sigProcess,
                           oldhandler=self._yolact_widget.process)
            self._yolact_dock.setVisible(False)
            self._seg_dock.setVisible(True)

    @qc.pyqtSlot(qw.QAction)
    def switchTracking(self, action):
        """Switch tracking between SORT and CSRT"""
        self._video_widget.pauseVideo()
        if action == self._sort_action:
            self._sort_dock.show()
            self._csrt_dock.hide()
            try:
                self._video_widget.sigSetFrame.disconnect(
                    self._csrt_widget.setFrame)
            except TypeError:
                pass
            newhandler = self._sort_widget.sigTrack
            oldhandler = self._csrt_widget.setBboxes
        else:  # csrt
            self._csrt_dock.show()
            self._sort_dock.hide()
            self._video_widget.sigSetFrame.connect(self._csrt_widget.setFrame)
            newhandler = self._csrt_widget.setBboxes
            oldhandler = self._sort_widget.sigTrack
        if self._yolact_action.isChecked():
            sig = self._lim_widget.sigProcessed
        else:
            sig = self._seg_widget.sigProcessed
        try:
            sig.disconnect()
        except TypeError:
            pass
        util.reconnect(sig, newhandler, oldhandler)

    @qc.pyqtSlot(str)
    def updateTitle(self, filename: str) -> None:
        self.setWindowTitle(f'Argos:track {filename}')

    @qc.pyqtSlot()
    def saveConfig(self):
        config = {
            'wmin': settings.value('segment/min_width', 10, type=int),
            'wmax': settings.value('segment/max_width', 50, type=int),
            'hmin': settings.value('segment/min_height', 10, type=int),
            'hmax': settings.value('segment/max_height', 100, type=int),
            'sort_metric': settings.value('sortracker/metric', 'iou', type=str),
            'min_dist': settings.value('sortracker/min_dist', 0.3, type=float),
            'min_hits': settings.value('sortracker/min_hits', 3, type=int),
            'max_age': settings.value('sortracker/max_age', 10, type=int),
            'pmin': settings.value('segment/min_pixels', 10, type=int),
            'pmax': settings.value('segment/max_pixels', 1000, type=int)
        }
        if self._seg_dock.isVisible():
            config['blur_width'] = settings.value('segment/blur_width', 21,
                                                  type=int)
            config['blur_sd'] = settings.value('segment/blur_sd', 1.0,
                                               type=float)
            config['thresh_method'] = settings.value('segment/thresh_method',
                                                     'gaussian', type=str)
            config['thresh_max'] = settings.value(
                'segment/thresh_max_intensity', 255, type=int)
            config['thresh_baseline'] = settings.value(
                'segment/thresh_baseline', type=int)
            config['thresh_blocksize'] = settings.value(
                'segment/thresh_blocksize', type=int)
            config['thresh_invert'] = settings.value('segment/thresh_invert',
                                                     True,
                                                     type=bool)
            seg_method = settings.value('segment/method', 'threshold',
                                        type=str).lower()
            config['seg_method'] = seg_method
            if seg_method == 'watershed':
                config['dist_thresh'] = int(settings.value(
                    'segment/watershed_distthresh', 3, type=float))
            elif seg_method == 'dbscan':
                config['eps'] = settings.value('segment/dbscan_eps', 5.0,
                                               type=float)
                config['min_samples'] = settings.value(
                    'segment/dbscan_minsamples', 10, type=int)
        else:  # yolact
            config['method'] = 'yolact'
            config['weight'] = settings.value('yolact/weightsfile', type=str)
            config['yconfig'] = settings.value('yolact/configfile', type=str)
            config['overlap_thresh'] = settings.value('yolact/overlap_thresh',
                                                      1.0, type=float)
            config['score_thresh'] = settings.value('yolact/score_thresh', 0.1,
                                                    type=float)
            config['top_k'] = settings.value('yolact/top_k', 30,
                                             type=int)
            config['cuda'] = settings.value('yolact/cuda', True, type=bool)

        directory = settings.value('video/directory', '.')
        fname, _ = qw.QFileDialog.getSaveFileName(
            self, 'Save configuration as',
            directory,
            filter='yaml (*.yml *.yaml)')
        if len(fname) > 0:
            with open(fname, 'w') as fd:
                yaml.dump(config, fd)

    @qc.pyqtSlot()
    def loadConfig(self):
        """TODO implement loading config file exported in yaml file"""
        raise NotImplementedError('This function is yet to be implemented')

    @qc.pyqtSlot(str)
    def statusMsgSlot(self, msg):
        self.statusBar().showMessage(msg)


if __name__ == '__main__':
    app = qw.QApplication(sys.argv)
    win = ArgosTracker()
    win.setMinimumSize(800, 600)
    win.setWindowTitle('Argos - track animals in video')
    win.showMaximized()
    app.aboutToQuit.connect(win.cleanup)
    win.show()
    sys.exit(app.exec_())
